import * as THREE from 'three';
import { Camera, Object3d } from './objects';
export class RaycasterEvent {
    type;
    cancelBubble;
    originalEvent;
    // Dummy default values
    coords = new THREE.Vector2(0, 0);
    distance = 0;
    intersected = false;
    constructor(type, originalEvent) {
        this.cancelBubble = false;
        this.type = type;
        this.originalEvent = originalEvent;
    }
    stopPropagation() {
        this.cancelBubble = true;
    }
}
export class RaycasterManager {
    raycaster;
    closestObject;
    pointer;
    canvas;
    camera;
    object3d;
    renderer;
    //currentObjectRaycaster: Object3d | null;
    constructor() {
        this.raycaster = new THREE.Raycaster();
        this.pointer = new THREE.Vector2(0, 0);
        this.object3d = [];
    }
    onCanvas = (canvas, renderer) => {
        this.canvas = canvas;
        this.renderer = renderer;
        window.addEventListener('click', this.onMouseClick);
        this.canvas.addEventListener('pointermove', this.onDocumentMouseMove);
        this.canvas.addEventListener('pointerdown', this.onMouseDown);
        this.canvas.addEventListener('pointerup', this.onMouseUp);
        this.canvas.addEventListener('mousemove', this.onDocumentMouseMove);
        window.addEventListener('mousedown', this.onMouseDown);
        window.addEventListener('mouseup', this.onMouseUp, true);
        this.canvas.addEventListener('touchstart', this.onTouchStart, {
            passive: true,
        });
        this.canvas.addEventListener('touchmove', this.onTouchMove, {
            passive: true,
        });
        this.canvas.addEventListener('touchend', this.onTouchEnd, {
            passive: true,
        });
    };
    dispose = () => {
        if (this.canvas != null) {
            window.removeEventListener('click', this.onMouseClick);
            this.canvas.ownerDocument.removeEventListener('pointermove', this.onDocumentMouseMove);
            this.canvas.removeEventListener('pointerdown', this.onMouseDown);
            this.canvas.removeEventListener('pointerup', this.onMouseUp);
            this.canvas.removeEventListener('mousemove', this.onDocumentMouseMove);
            window.removeEventListener('mousedown', this.onMouseDown);
            window.removeEventListener('mouseup', this.onMouseUp);
            this.canvas.removeEventListener('touchstart', this.onTouchStart);
            this.canvas.removeEventListener('touchmove', this.onTouchMove);
            this.canvas.removeEventListener('touchend', this.onTouchEnd);
        }
    };
    add = (object) => {
        if (object) {
            let newObject = new Object3d(object);
            this.object3d.push(newObject);
            return newObject;
        }
    };
    update = (pointer, camera) => {
        //console.log (pointer)
        this.pointer = pointer;
        this.camera = camera;
        let direcition = this.camera.getWorldDirection;
        this.raycaster.setFromCamera(pointer, camera);
        if (this.object3d != null) {
            this.object3d.forEach((object) => {
                if (object.target)
                    this.checkIntersection(object);
            });
        }
        this.object3d.sort(function (a, b) {
            return a.distance - b.distance;
        });
        const newClosestObject = this.object3d.find((object) => object.intersected) ?? null;
        if (newClosestObject != this.closestObject) {
            if (this.closestObject) {
                const eventOutClosest = new RaycasterEvent('mouseout');
                this.dispatch(this.closestObject, eventOutClosest);
            }
            if (newClosestObject) {
                const eventOverClosest = new RaycasterEvent('mouseover');
                this.dispatch(newClosestObject, eventOverClosest);
            }
            this.closestObject = newClosestObject;
        }
    };
    checkIntersection = (object) => {
        const intersects = this.raycaster.intersectObjects([object.target], true);
        object.wasIntersected = object.intersected;
        if (intersects.length > 0) {
            let distance = intersects[0].distance;
            intersects.forEach((i) => {
                if (i.distance < distance) {
                    distance = i.distance;
                }
                object.intersected = true;
                object.distance = distance;
                if (object.callback) {
                    object.callback();
                }
            });
        }
        else {
            object.intersected = false;
        }
    };
    dispatch = (object, event) => {
        if (object.target && !event.cancelBubble) {
            event.coords = this.pointer;
            event.distance = object.distance;
            event.intersected = object.intersected;
            object.target.dispatchEvent(event);
        }
    };
    onMouseClick = (mouseEvent) => {
        const event = new RaycasterEvent('click', mouseEvent);
        if (this.object3d != null) {
            this.object3d.forEach((object) => {
                if (object.intersected) {
                    this.dispatch(object, event);
                }
            });
        }
    };
    onMouseDown = (mouseEvent) => {
        if (this.camera)
            this.update(this.pointer, this.camera);
        const event = new RaycasterEvent('mousedown', mouseEvent);
        if (this.object3d != null) {
            this.object3d.forEach((object) => {
                if (object.intersected) {
                    this.dispatch(object, event);
                }
            });
        }
    };
    onMouseUp = (mouseEvent) => {
        const event = new RaycasterEvent('mouseup', mouseEvent);
        if (this.object3d != null) {
            this.object3d.forEach((object) => {
                this.dispatch(object, event);
            });
        }
    };
    onDocumentMouseMove = (mouseEvent) => {
        const event = new RaycasterEvent('mousemove', mouseEvent);
        if (this.object3d != null) {
            this.object3d.forEach((object) => {
                if (object.intersected) {
                    this.dispatch(object, event);
                }
            });
        }
    };
    onTouchMove = (touchEvent) => {
        //console.log("onTouchMove");
    };
    onTouchStart = (touchEvent) => {
        //console.log("onTouchStart");
    };
    onTouchEnd = (touchEvent) => {
        //console.log("onTouchEnd");
    };
}
